/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file recv_multipart.c
 * Implementation of the s3 DSI Receive
 */
#include "globus_gridftp_server_s3_base.h"

static
void *
globus_l_s3_recv_multipart_thread(
    void                               *user_arg);

static
S3Status
globus_l_s3_recv_multipart_initiate_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data);

static
void
globus_l_s3_recv_multipart_initiate_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_initiate_response_callback(
    const char                         *upload_id,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_upload_part_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data);

static
void
globus_l_s3_recv_multipart_upload_part_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
int
globus_l_s3_recv_multipart_upload_part_data(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_commit_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data);

static
void
globus_l_s3_recv_multipart_commit_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);
static
int
globus_l_s3_recv_multipart_commit_put_data(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_commit_response(
    const char                         *location,
    const char                         *etag,
    void                               *callback_data);

static
globus_result_t
globus_l_s3_recv_multipart_initiate(
    globus_i_s3_transfer_t             *transfer_state);

static
globus_result_t
globus_l_s3_recv_multipart_register_reads(
    globus_i_s3_transfer_t             *transfer_state);

static
void
globus_l_s3_recv_multipart_read_callback(
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t                      *buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void                               *user_arg);

static
globus_result_t
globus_l_s3_recv_multipart_initialize_parts(
    globus_i_s3_transfer_t             *transfer_state);

globus_result_t
globus_i_s3_recv_multipart(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    uint64_t                            transfer_size,
    uint64_t                            part_size,
    bool                                restart,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_i_s3_transfer_t             *transfer_state = NULL;
    globus_thread_t                     t;
    int                                 rc;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    assert(transfer_info != NULL);
    assert(s3_handle != NULL);

    transfer_state = malloc(sizeof(globus_i_s3_transfer_t) + strlen(transfer_info->pathname) + 1);
    if (transfer_state == NULL)
    {
        result = GlobusGFSErrorMemory("transfer_state");
        goto fail;
    }

    *transfer_state = (globus_i_s3_transfer_t)
    {
        .op                           = op,
        .transfer_info                = transfer_info,
        .pending_buffers_last         = &transfer_state->pending_buffers,
        .total_file_size              = transfer_size,
        .part_size                    = part_size,
        .multipart_restart            = restart
    };

    globus_mutex_init(&transfer_state->mutex, NULL);
    globus_cond_init(&transfer_state->cond, NULL);

    strcpy(transfer_state->pathname, transfer_info->pathname);
    globus_i_s3_path_parse(transfer_state->pathname, &transfer_state->bucket_name, &transfer_state->key);

    transfer_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, transfer_state->bucket_name);

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_recv_multipart_thread,
            transfer_state);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");

        globus_cond_destroy(&transfer_state->cond);
        globus_mutex_destroy(&transfer_state->mutex);

        if (transfer_state->parts != NULL)
        {
            free(transfer_state->parts);
        }

fail:
        GlobusS3Info("finished_transfer: op=%p result=%#x\n",
                    (void *) op,
                    (unsigned int) result);
        
        globus_gridftp_server_finished_transfer(op, result);
        free(transfer_state);
    }

    return result;
}
/* globus_i_s3_recv_multipart() */

static
void *
globus_l_s3_recv_multipart_thread(
    void                               *user_arg)
{
    globus_i_s3_transfer_t             *transfer_state = user_arg;
    globus_result_t                     result;
    globus_off_t                        offset, length;
    size_t                              content_length = 0;

    GlobusS3Enter();

    globus_mutex_lock(&transfer_state->mutex);

    // List multipart uploads for this key and find the earliest-created one
    // This also removes all other partial uploads for this key.
    result = globus_i_s3_recv_multipart_resolve_id(transfer_state);

    if (transfer_state->multipart_restart)
    {
        if ((result == GLOBUS_SUCCESS)
            && (transfer_state->multipart_id == NULL))
        {
            result = GlobusGFSErrorGeneric("No multi-part upload found for restart");
        }

        if (result != GLOBUS_SUCCESS)
        {
            goto failed_restart;
        }
    }
    else
    {
        if (transfer_state->multipart_id != NULL)
        {
            // If not restarting, and there's an already-existing partial
            // upload, we remove it and clear it from the state object
            globus_i_s3_recv_multipart_abort(transfer_state);

            free(transfer_state->multipart_id);
            transfer_state->multipart_id = NULL;
        }

        result = globus_l_s3_recv_multipart_initiate(transfer_state);
    }

    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        result = transfer_state->result;
    }
    else if (transfer_state->multipart_id == NULL)
    {
        result = GlobusGFSErrorGeneric("Cannot determine upload ID for transfer");
    }
    if (result != GLOBUS_SUCCESS)
    {
        goto failed_initiate;
    }

    // In a restart, we don't know the length of the file, but will
    // make the following assumptions about the client in order to 
    // handle the "normal case" of an interrupted client resuming
    // based on the restart markers we sent
    //
    // 1) the transfer data will come to us in order
    // 2) the client will not attempt to send data that doesn't align
    //    with the restart markers we sent.
    // 3) the client sends an ALLO message with the amount of data being
    //    sent in this transfer, and that amount corresponds to "the
    //    rest of the file" since the last restart marker
    //

    // Get the transfer write range. This will do magic so that offsets
    // in the read callbacks are legitimate. We also use the sum of the 
    // offset and the alloc_size as the total size of the file being
    // transferred
    globus_gridftp_server_get_write_range(
            transfer_state->op,
            &offset,
            &length);

    if (length != (globus_off_t) -1)
    {
        result = GlobusGFSErrorGeneric("Cannot determine write range");
        goto failed_initiate;
    }
    transfer_state->offset = offset;
    if (transfer_state->multipart_restart)
    {
        transfer_state->total_file_size = 
                    offset + transfer_state->transfer_info->alloc_size;
    }

    GlobusS3Info("begin_transfer: op=%p pathname=%s\n",
                 (void *) transfer_state->op, transfer_state->transfer_info->pathname);
    globus_gridftp_server_begin_transfer(transfer_state->op, 0, NULL);

    content_length = snprintf(NULL, 0, "<CompleteMultipartUpload></CompleteMultipartUpload>");

    // Initialize transfer_state->num_parts and transfer_state->parts
    result = globus_l_s3_recv_multipart_initialize_parts(transfer_state);

    if (result != GLOBUS_SUCCESS)
    {
        goto failed_initialize_parts;
    }

    if (transfer_state->multipart_restart)
    {
        // List the parts of the transfer to get an idea about what has been
        // received already. This will set the value of etags that the
        // upload already contains
        result = globus_i_s3_recv_multipart_list_parts(transfer_state);

        if (result != GLOBUS_SUCCESS)
        {
            goto failed_restart;
        }
    }

    // Read the part data from the data channel and send it to the s3 service
    for (size_t i = 0; transfer_state->result == GLOBUS_SUCCESS && i < transfer_state->num_parts; i++)
    {
        transfer_state->retries = 0;

        if (transfer_state->parts[i].etag[0] == 0)
        {
            do 
            {
                transfer_state->status = S3StatusOK;
                transfer_state->result = GLOBUS_SUCCESS;

                globus_i_s3_retry_delay(transfer_state->retries);

                GlobusS3Info("REST: op=%p S3_upload_part(id=%s, part=%zu) try=%d\n",
                            (void *) transfer_state->op,
                            transfer_state->multipart_id,
                            i+1,
                            transfer_state->retries);

                S3_upload_part(
                    &transfer_state->bucket_context,
                    transfer_state->key,
                    NULL,
                    &(S3PutObjectHandler)
                    {
                        .responseHandler =
                        {
                            .propertiesCallback
                                          = globus_l_s3_recv_multipart_upload_part_properties,
                            .completeCallback
                                          = globus_l_s3_recv_multipart_upload_part_complete,
                        },
                        .putObjectDataCallback
                                          = globus_l_s3_recv_multipart_upload_part_data
                    },
                    i+1,
                    transfer_state->multipart_id,
                    (int) transfer_state->parts[i].part_length,
                    NULL,
                    transfer_state);
                GlobusS3DebugOpStatus(S3_upload_part, transfer_state);
            }
            while (S3_status_is_retryable(transfer_state->status)
                    && (transfer_state->retries++) < globus_i_s3_max_retries);

            if (transfer_state->result != GLOBUS_SUCCESS)
            {
                break;
            }
            GlobusS3Debug("restart_marker: op=%p offset=%"PRIuMAX" length=%"PRIuMAX"\n",
                        (void *) transfer_state->op,
                        (uintmax_t) transfer_state->parts[i].part_offset,
                        (uintmax_t) transfer_state->parts[i].part_length);

            globus_gridftp_server_update_range_recvd(
                    transfer_state->op,
                    transfer_state->parts[i].part_offset,
                    transfer_state->parts[i].part_length);
        }
        content_length += snprintf(NULL, 0, "<Part><PartNumber>%zd</PartNumber><ETag>%s</ETag></Part>",
                i+1, transfer_state->parts[i].etag);

    }
    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        result = transfer_state->result;
        goto failed_upload;
    }
    /* Possible that libs3 called finished before we registered a 
     * callback that resulted in EOF.
     */
    if (transfer_state->registered_buffers == NULL && ! transfer_state->eof)
    {
        globus_l_s3_recv_multipart_register_reads(transfer_state);
    }

    transfer_state->retries = 0;
    do
    {
        transfer_state->status = S3StatusOK;
        transfer_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(transfer_state->retries);

        GlobusS3Info("REST: op=%p S3_complete_multipart_upload(id=%s) try=%d\n",
                     (void *) transfer_state->op,
                     transfer_state->multipart_id,
                     transfer_state->retries);

        S3_complete_multipart_upload(
                &transfer_state->bucket_context,
                transfer_state->key,
                &(S3MultipartCommitHandler)
                {
                    .responseHandler.propertiesCallback 
                                          = globus_l_s3_recv_multipart_commit_properties,
                    .responseHandler.completeCallback
                                          = globus_l_s3_recv_multipart_commit_complete,
                    .putObjectDataCallback
                                          = globus_l_s3_recv_multipart_commit_put_data,
                    .responseXmlCallback  = globus_l_s3_recv_multipart_commit_response
                }, 
                transfer_state->multipart_id,
                (int) content_length, 
                NULL,
                transfer_state);
        GlobusS3DebugOpStatus(S3_complete_multipart_upload, transfer_state);
    }
    while (S3_status_is_retryable(transfer_state->status)
        && (transfer_state->retries++) < globus_i_s3_max_retries);


    result = transfer_state->result;

failed_initiate:
failed_upload:
failed_initialize_parts:
failed_restart:
    GlobusS3Info("finished_transfer: op=%p result=%#x\n",
                 (void *) transfer_state->op,
                 (unsigned int) result);

    globus_gridftp_server_finished_transfer(transfer_state->op, result);

    while (transfer_state->registered_buffers != NULL)
    {
        globus_cond_wait(&transfer_state->cond, &transfer_state->mutex);
    }

    while (transfer_state->pending_buffers != NULL)
    {
        globus_i_s3_buffer_t           *buffer = transfer_state->pending_buffers;
        transfer_state->pending_buffers = transfer_state->pending_buffers->next;

        free(buffer);
    }
    globus_mutex_unlock(&transfer_state->mutex);
    globus_mutex_destroy(&transfer_state->mutex);
    globus_cond_destroy(&transfer_state->cond);

    free(transfer_state->parts);
    free(transfer_state->multipart_id);
    free(transfer_state);

    return NULL;
}
/* globus_l_s3_recv_multipart_thread() */

static
S3Status
globus_l_s3_recv_multipart_initiate_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_initiate_properties() */

static
void
globus_l_s3_recv_multipart_initiate_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    assert(transfer_state != NULL);

    if (status != S3StatusOK && transfer_state->result == GLOBUS_SUCCESS)
    {
        transfer_state->status = status;
        transfer_state->result = GlobusGFSErrorS3(status, transfer_state->transfer_info->pathname, error_details);
    }

    GlobusS3Exit();

    return;
}
/* globus_l_s3_recv_multipart_initiate_complete() */

static
S3Status
globus_l_s3_recv_multipart_initiate_response_callback(
    const char                         *upload_id,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    if (upload_id)
    {
        GlobusS3Debug("transfer_id=%s\n", upload_id);

        transfer_state->multipart_id = strdup(upload_id);
    }
    if (transfer_state->multipart_id == NULL)
    {
        if (transfer_state->result == GLOBUS_SUCCESS)
        {
            transfer_state->result = GlobusGFSErrorGeneric("No multipart upload id");
            status = S3StatusInternalError;
        }
    }

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_initial_response_callback() */

static
S3Status
globus_l_s3_recv_multipart_upload_part_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    for (size_t i = 0; i < transfer_state->num_parts; i++)
    {
        if (transfer_state->parts[i].etag[0] == 0)
        {
            size_t len;

            GlobusS3Debug("part=%zd etag=%s\n",
                i+1,
                properties->eTag);

            assert(properties->eTag != NULL);
            len = snprintf(
                    transfer_state->parts[i].etag,
                    sizeof(transfer_state->parts[i].etag),
                    "%s",
                    properties->eTag);

            assert(len == strlen(properties->eTag));
            assert(len < sizeof(transfer_state->parts[i].etag));

            break;
        }
    }

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_upload_part_properties() */

static
void
globus_l_s3_recv_multipart_upload_part_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && transfer_state->result == GLOBUS_SUCCESS)
    {
        transfer_state->status = S3StatusOK;
        transfer_state->result = GlobusGFSErrorS3(status, transfer_state->transfer_info->pathname, error_details);
    }

    GlobusS3Exit();

    return;
}
/* globus_l_s3_recv_multipart_upload_part_complete() */

static
int
globus_l_s3_recv_multipart_upload_part_data(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;
    int                                 data_copied = 0;
    int                                 end_of_part = -1;

    GlobusS3Enter();

    while ((!globus_i_s3_recv_is_transfer_offset_ready(transfer_state))
           && (!globus_i_s3_recv_is_reading_complete(transfer_state)))
    {
        GlobusS3Trace("Waiting for current offset %"PRIuMAX" to be ready\n",
                      (uintmax_t) transfer_state->offset);

        globus_l_s3_recv_multipart_register_reads(transfer_state);

        globus_cond_wait(&transfer_state->cond, &transfer_state->mutex);
    }

    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        data_copied = -1;

        goto done;
    }
    for (size_t i = 0; i < transfer_state->num_parts; i++)
    {
        if (transfer_state->offset >= transfer_state->parts[i].part_offset)
        {
            end_of_part = transfer_state->parts[i].part_offset + transfer_state->parts[i].part_length;
            if (transfer_state->offset < end_of_part)
            {
                break;
            }
        }
    }

    while ((data_copied < buffer_size)
        && globus_i_s3_recv_is_transfer_offset_ready(transfer_state))
    {
        int                             to_copy;
        globus_i_s3_buffer_t           *s3_buffer;

        s3_buffer = transfer_state->pending_buffers;
        to_copy = buffer_size - data_copied;

        if (to_copy > s3_buffer->buffer_used)
        {
            to_copy = s3_buffer->buffer_used;
        }
        if ((s3_buffer->transfer_offset + to_copy) > (end_of_part))
        {
            to_copy = end_of_part - (s3_buffer->transfer_offset + to_copy);
        }

        if (to_copy == 0)
        {
            break;
        }

        memcpy(buffer+data_copied, s3_buffer->buffer, to_copy);
        data_copied += to_copy;
        if (to_copy < s3_buffer->buffer_used)
        {
            memmove(s3_buffer->buffer, s3_buffer->buffer + to_copy, s3_buffer->buffer_used - to_copy);
            s3_buffer->buffer_used -= to_copy;
            s3_buffer->transfer_offset += to_copy;
        }
        else
        {
            transfer_state->pending_buffers = s3_buffer->next;
            if (transfer_state->pending_buffers == NULL)
            {
                transfer_state->pending_buffers_last = &transfer_state->pending_buffers;
            }
            globus_i_s3_buffer_release(transfer_state, s3_buffer);
        }
        transfer_state->offset += to_copy;
    }

done:
    GlobusS3ExitInt(data_copied);

    return data_copied;
}
/* globus_l_s3_recv_multipart_upload_part_data() */

static
globus_result_t
globus_l_s3_recv_multipart_register_reads(
    globus_i_s3_transfer_t             *transfer_state)
{
    globus_i_s3_buffer_t               *buffer;
    globus_size_t                       optimal_blocksize;
    int                                 optimal_concurrency;
    int                                 currently_registered;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    buffer = transfer_state->registered_buffers;
    currently_registered = 0;

    while (buffer != NULL)
    {
        currently_registered++;
        buffer = buffer->next;
    }

    if (transfer_state->result == GLOBUS_SUCCESS && !transfer_state->eof)
    {
        /* Reads still might be useful */
        globus_gridftp_server_get_block_size(
                transfer_state->op,
                &optimal_blocksize);

        globus_gridftp_server_get_optimal_concurrency(
                transfer_state->op,
                &optimal_concurrency);
        
        for (int i = currently_registered; i < optimal_concurrency; i++)
        {
            buffer = globus_i_s3_buffer_get(
                    transfer_state, (size_t) optimal_blocksize);
            transfer_state->current_buffer = NULL;

            if (buffer == NULL)
            {
                result = GlobusGFSErrorMemory("buffer");

                goto buffer_fail;
            }
            buffer->transfer_offset = (uint64_t) -1;

            assert(buffer->buffer_used == 0);

            result = globus_gridftp_server_register_read(
                    transfer_state->op,
                    buffer->buffer,
                    buffer->buffer_len,
                    globus_l_s3_recv_multipart_read_callback,
                    transfer_state);
            if (result != GLOBUS_SUCCESS)
            {
                goto register_fail;
            }
            buffer->next = transfer_state->registered_buffers;
            transfer_state->registered_buffers = buffer;
        }
    }

register_fail:
buffer_fail:
    GlobusS3Exit();
    return result;
}
/* globus_l_s3_recv_register_reads() */

static
void
globus_l_s3_recv_multipart_read_callback(
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t                      *buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void                               *user_arg)
{
    globus_i_s3_transfer_t             *transfer_state = user_arg;
    globus_i_s3_buffer_t               *s3_buffer;
    globus_i_s3_buffer_t               *tmp;
    globus_i_s3_buffer_t              **prev_next;
    bool                                signal = eof;

    GlobusS3Enter();

    GlobusS3Debug("recv_read_callback: op=%p result=%#x buffer=%p nbytes=%zd offset=%"PRIuMAX" eof=%d user_arg=%p",
            (void *) op,
            result,
            (void *) buffer,
            (size_t) nbytes,
            (uintmax_t) offset,
            (int) eof,
            user_arg);

    globus_mutex_lock(&transfer_state->mutex);

    transfer_state->eof |= eof;

    /* Remove from the registered buffer list */
    s3_buffer = transfer_state->registered_buffers;
    prev_next = &transfer_state->registered_buffers;

    while (s3_buffer != NULL)
    { 
        if (s3_buffer->buffer == buffer)
        {
            *prev_next = s3_buffer->next;
            break;
        }
        prev_next = &s3_buffer->next;
        s3_buffer = s3_buffer->next;
    }

    assert(s3_buffer != NULL);

    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        transfer_state->eof = true;
        goto bad_read;
    }

    /* Update the info about this buffer */
    s3_buffer->transfer_offset = offset;
    s3_buffer->buffer_used = nbytes;

    globus_gridftp_server_update_bytes_recvd(
            transfer_state->op,
            nbytes);

    /*
     * If something bad occurred, we will discard this block and flag
     * the request with an error
     */
    if (result != GLOBUS_SUCCESS)
    {
        transfer_state->result = result;
        transfer_state->eof = true;
        goto bad_read;
    }

    if (transfer_state->total_file_size != (uint64_t) -1)
    {
        if ((uint64_t) (offset + nbytes) > 
            (uint64_t) transfer_state->total_file_size)
        {
            /* peer sent me more than I told I'd send to s3. error! */
            transfer_state->result = GlobusGFSErrorSystemError("write", EFBIG);
            transfer_state->eof = true;
            goto bad_read;
        }
    }

    if (s3_buffer->buffer_used == 0)
    {
        /* empty buffer */
        s3_buffer->next = transfer_state->free_buffers;
        transfer_state->free_buffers = s3_buffer;
    }
    else
    {
        /* Add it to the pending buffer list (in order) */
        tmp = transfer_state->pending_buffers;
        prev_next = &transfer_state->pending_buffers;
        while (tmp != NULL)
        {
            if ((tmp->transfer_offset + tmp->buffer_used) <= s3_buffer->transfer_offset)
            {
                prev_next = &tmp->next;
                tmp = tmp->next;
            }
            else
            {
                break;
            }
        }
        s3_buffer->next = tmp;
        if (tmp == NULL)
        {
            transfer_state->pending_buffers_last = &s3_buffer->next;
        }

        *prev_next = s3_buffer;

        signal = true;
    }

    if (signal)
    {
        globus_cond_signal(&transfer_state->cond);
    }
    globus_mutex_unlock(&transfer_state->mutex);

    GlobusS3Exit();
    return;

bad_read:
    /* Since this was a bad read, we'll ignore the data */
    s3_buffer->buffer_used = 0;
    s3_buffer->transfer_offset = 0;
    s3_buffer->next = transfer_state->free_buffers;
    transfer_state->free_buffers = s3_buffer;

    globus_cond_signal(&transfer_state->cond);
    globus_mutex_unlock(&transfer_state->mutex);

    GlobusS3Exit();
    return;
}
/* globus_l_s3_recv_multipart_read_callback() */

static
S3Status
globus_l_s3_recv_multipart_commit_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();
    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_commit_properties() */

static
void
globus_l_s3_recv_multipart_commit_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && transfer_state->result == GLOBUS_SUCCESS)
    {
        transfer_state->result = GlobusGFSErrorS3(status, transfer_state->transfer_info->pathname, error_details);
    }
    GlobusS3Exit();

    return;
}
/* globus_l_s3_recv_multipart_commit_complete() */

static
int
globus_l_s3_recv_multipart_commit_put_data(
    int                                 buffer_size,
    char                               *buffer,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;
    int                                 copied = 0;
    bool                                parts_done = false;

    GlobusS3Enter();

    buffer[0] = 0;

    if (transfer_state->complete_end)
    {
        goto done;
    }
    if (transfer_state->parts[0].etag[0])
    {
        assert(buffer_size > strlen("<CompleteMultipartUpload>"));

        copied += snprintf(buffer+copied, buffer_size-copied, "<CompleteMultipartUpload>");
    }
    for (size_t i = 0; i < transfer_state->num_parts; i++)
    {
        if (transfer_state->parts[i].etag[0] == 0)
        {
            if (i == (transfer_state->num_parts - 1))
            {
                parts_done = true;
            }
            continue;
        }
        else
        {
            size_t                          part_frag_length;

            part_frag_length = snprintf(
                    NULL,
                    0,
                    "<Part><PartNumber>%d</PartNumber><ETag>%s</ETag></Part>",
                    (int) i+1,
                    transfer_state->parts[i].etag);

            if ((copied + part_frag_length + 1) < buffer_size)
            {
                copied += snprintf(buffer+copied, buffer_size-copied, "<Part><PartNumber>%d</PartNumber><ETag>%s</ETag></Part>", (int) i+1, transfer_state->parts[i].etag);
                transfer_state->parts[i].etag[0] = 0;
                if (i == (transfer_state->num_parts-1))
                {
                    parts_done = true;
                }
            }
            else
            {
                break;
            }
        }
    }
    if (parts_done)
    {
        static const char complete_frag[]="</CompleteMultipartUpload>";

        if ((copied + sizeof(complete_frag) - 1) <= buffer_size)
        {
            memcpy(buffer+copied, complete_frag, sizeof(complete_frag)-1);
            copied += sizeof(complete_frag) - 1;
            transfer_state->complete_end = true;
        }
    }

done:

    GlobusS3Debug("buffer=\"%.*s\"\n", copied, buffer);

    GlobusS3ExitInt(copied);

    return copied;
}
/* globus_l_s3_recv_multipart_commit_put_data() */

static
S3Status
globus_l_s3_recv_multipart_commit_response(
    const char                         *location,
    const char                         *etag,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);
    
    return status;
}
/* globus_l_s3_recv_multipart_commit_response() */

static
globus_result_t
globus_l_s3_recv_multipart_initiate(
    globus_i_s3_transfer_t             *transfer_state)
{
    char                                s3_md5[((MD5_DIGEST_LENGTH+2)*4)/3+1];
    S3PutProperties                     put_properties;
    time_t                              modification_time;
    char                                modification_time_string[snprintf(NULL, 0, "%"PRIdMAX, (intmax_t) INTMAX_MAX)+2];
    S3NameValue                         meta_data[2];
    globus_result_t                     result = GLOBUS_SUCCESS;

    // Initiate the multipart transfer with the metadata from the user
    put_properties = (S3PutProperties)
    {
        .contentType              = globus_i_s3_mime_type_get(transfer_state->key),
        .metaData                 = meta_data,
        .metaDataCount            = 0
    };

    result = globus_gridftp_server_get_recv_modification_time(
            transfer_state->op,
            &modification_time);
    if (result == GLOBUS_SUCCESS && modification_time != (time_t) -1)
    {
        snprintf(modification_time_string,
                 sizeof(modification_time_string),
                 "%"PRIdMAX,
                (intmax_t) modification_time);

        meta_data[put_properties.metaDataCount++] = (S3NameValue)
        {
            .name                 = GLOBUS_S3_MTIME_NAME,
            .value                = modification_time_string
        };
    }

    if (transfer_state->transfer_info->expected_checksum_alg != NULL
        && strcasecmp(transfer_state->transfer_info->expected_checksum_alg, "md5") == 0)
    {
        if (strlen(transfer_state->transfer_info->expected_checksum) != 32)
        {
            result = GlobusGFSErrorGeneric("Invalid md5 checksum");
            goto bad_checksum;
        }

        result = globus_i_s3_md5sum_encode(
                transfer_state->transfer_info->expected_checksum, s3_md5);
        if (result != GLOBUS_SUCCESS)
        {
            goto bad_checksum;
        }
        meta_data[put_properties.metaDataCount++] = (S3NameValue)
        {
            .name                 = GLOBUS_S3_MD5_NAME,
            .value                = s3_md5
        };
        put_properties.md5 = s3_md5;
    }

    transfer_state->retries = 0;
    do 
    {
        transfer_state->status = S3StatusOK;
        transfer_state->result = GLOBUS_SUCCESS;

        globus_i_s3_retry_delay(transfer_state->retries);

        GlobusS3Info("REST: op=%p S3_initiate_multipart(bucket=%s, key=%s) try=%d\n",
                    (void *) transfer_state->op,
                    transfer_state->bucket_name,
                    transfer_state->key,
                    transfer_state->retries);
        S3_initiate_multipart(
            &transfer_state->bucket_context,
            transfer_state->key,
            &put_properties,
            &(S3MultipartInitialHandler)
            {
                .responseHandler.propertiesCallback
                                      = globus_l_s3_recv_multipart_initiate_properties,
                .responseHandler.completeCallback
                                      = globus_l_s3_recv_multipart_initiate_complete,
                .responseXmlCallback  = globus_l_s3_recv_multipart_initiate_response_callback
            },
            NULL,
            transfer_state);
        GlobusS3DebugOpStatus(S3_initiate_multipart, transfer_state);
    }
    while (S3_status_is_retryable(transfer_state->status)
            && (transfer_state->retries++) < globus_i_s3_max_retries);

bad_checksum:
    if (transfer_state->result != GLOBUS_SUCCESS)
    {
        result = transfer_state->result;
    }

    GlobusS3ExitResult(result);

    return result;
}
/* globus_l_s3_recv_multipart_initiate() */

/**
 * @brief Initialize the num_parts and parts fields of the transfer state
 * structure
 */
static
globus_result_t
globus_l_s3_recv_multipart_initialize_parts(
    globus_i_s3_transfer_t             *transfer_state)
{
    uint64_t                            final_part_size;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    final_part_size = transfer_state->total_file_size % transfer_state->part_size;

    transfer_state->num_parts = transfer_state->total_file_size / transfer_state->part_size
            + (final_part_size != 0);
    if (final_part_size == 0)
    {
        final_part_size = transfer_state->part_size;
    }

    transfer_state->parts = malloc(sizeof(globus_i_s3_part_t) * transfer_state->num_parts);
    if (transfer_state->parts == NULL)
    {
        result = GlobusGFSErrorMemory("parts");

        goto parts_malloc_fail;
    }

    for (size_t i = 0; i < transfer_state->num_parts; i++)
    {
        transfer_state->parts[i] = (globus_i_s3_part_t)
        {
            .etag                     = {0},
            .part_offset              = i * transfer_state->part_size,
            .part_length              = (i == (transfer_state->num_parts - 1))
                                        ? final_part_size : transfer_state->part_size
        };
    }

parts_malloc_fail:
    GlobusS3ExitResult(result);
    return result;
}
/* globus_l_s3_recv_multipart_initialize_parts() */
// vim: ft=c:
